
//******************************************************************************
// RCF - Remote Call Framework
// Copyright (c) 2005 - 2008, Jarl Lindrud. All rights reserved.
// Consult your license for conditions of use.
// Version: 0.9d
// Contact: jarl.lindrud <at> gmail.com 
//******************************************************************************

#include <RCF/RcfSession.hpp>
#include <RCF/SerializationProtocol.hpp>
#include <RCF/Version.hpp>

#include <boost/bind.hpp>

namespace RCF {

    RcfSession::RcfSession() :
        mFiltered(RCF_DEFAULT_INIT),
        mStopCallInProgress(RCF_DEFAULT_INIT),
        mRcfRuntimeVersion(RCF::getRuntimeVersion()),
        mTransportFiltersLocked(RCF_DEFAULT_INIT)
    {}

    RcfSession::~RcfSession()
    {
        RCF_DTOR_BEGIN
            // no locks here, relying on dtor thread safety of reference counted objects
            if (mOnDestroyCallback)
            {
                mOnDestroyCallback(*this);
            }
        RCF_DTOR_END
    }

    void RcfSession::setOnDestroyCallback(OnDestroyCallback onDestroyCallback)
    {
        Lock lock(mMutex);
        mOnDestroyCallback = onDestroyCallback;
    }

#ifdef RCF_USE_SF_SERIALIZATION

    void RcfSession::enableSfSerializationPointerTracking(bool enable)
    {
        mOut.mOutProtocol1.setCustomizationCallback(
            boost::bind(enableSfPointerTracking_1, _1, enable) );

        //mOut.mOutProtocol2.setCustomizationCallback(
        //    boost::bind(enableSfPointerTracking_2, _1, enable) );
    }

#else

    void RcfSession::enableSfSerializationPointerTracking(bool enable)
    {}

#endif

    void RcfSession::addOnWriteCompletedCallback(
        const OnWriteCompletedCallback &onWriteCompletedCallback)
    {
        Lock lock(mMutex);
        mOnWriteCompletedCallbacks.push_back(onWriteCompletedCallback);
    }

    void RcfSession::addOnWriteInitiatedCallback(
        const OnWriteInitiatedCallback &onWriteInitiatedCallback)
    {
        Lock lock(mMutex);
        mOnWriteInitiatedCallbacks.push_back(onWriteInitiatedCallback);
    }

    void RcfSession::extractOnWriteCompletedCallbacks(
        std::vector<OnWriteCompletedCallback> &onWriteCompletedCallbacks)
    {
        Lock lock(mMutex);
        onWriteCompletedCallbacks.clear();
        onWriteCompletedCallbacks.swap( mOnWriteCompletedCallbacks );
    }

    void RcfSession::extractOnWriteInitiatedCallbacks(
        std::vector<OnWriteInitiatedCallback> &onWriteInitiatedCallbacks)
    {
        Lock lock(mMutex);
        onWriteInitiatedCallbacks.clear();
        onWriteInitiatedCallbacks.swap( mOnWriteInitiatedCallbacks );
    }

    const RCF::I_RemoteAddress &RcfSession::getRemoteAddress()
    {
        return getProactorPtr()->getRemoteAddress();
    }

    void RcfSession::disconnect()
    {
        dynamic_cast<RCF::I_ServerTransportEx & >(
            getProactorPtr()->getServerTransport())
                .createClientTransport(RCF::getCurrentRcfSessionPtr());
    }

    bool RcfSession::hasDefaultServerStub()
    {
        Lock lock(mMutex);
        return mDefaultStubEntryPtr;
    }

    StubEntryPtr RcfSession::getDefaultStubEntryPtr()
    {
        Lock lock(mMutex);
        return mDefaultStubEntryPtr;
    }

    void RcfSession::setDefaultStubEntryPtr(const StubEntryPtr &stubEntryPtr)
    {
        Lock lock(mMutex);
        mDefaultStubEntryPtr = stubEntryPtr;
    }

    void RcfSession::setCachedStubEntryPtr(const StubEntryPtr &stubEntryPtr)
    {
        mCachedStubEntryPtr = stubEntryPtr;
    }

    void RcfSession::getMessageFilters(std::vector<FilterPtr> &filters)
    {
        filters = mFilters;
    }

    void RcfSession::getTransportFilters(std::vector<FilterPtr> &filters)
    {
        getProactorPtr()->getTransportFilters(filters);
    }

    int RcfSession::getRcfRuntimeVersion()
    {
        return mRcfRuntimeVersion;
    }

    void RcfSession::setRcfRuntimeVersion(int version)
    {
        mRcfRuntimeVersion = version;
    }

    void RcfSession::setUserData(boost::any userData)
    {
        mUserData = userData;
    }

    boost::any RcfSession::getUserData()
    {
        return mUserData;
    }

    void RcfSession::lockTransportFilters()
    {
        mTransportFiltersLocked = true;
    }

    void RcfSession::unlockTransportFilters()
    {
        mTransportFiltersLocked = false;
    }

    bool RcfSession::transportFiltersLocked()
    {
        return mTransportFiltersLocked;
    }

} // namespace RCF
