
//******************************************************************************
// RCF - Remote Call Framework
// Copyright (c) 2005 - 2008, Jarl Lindrud. All rights reserved.
// Consult your license for conditions of use.
// Version: 0.9d
// Contact: jarl.lindrud <at> gmail.com 
//******************************************************************************

#include <RCF/TcpAsioServerTransport.hpp>

#include <boost/bind.hpp>
#include <boost/shared_ptr.hpp>

#include <RCF/AsyncFilter.hpp>
#include <RCF/CurrentSession.hpp>
#include <RCF/MethodInvocation.hpp>
#include <RCF/RcfServer.hpp>
#include <RCF/TcpClientTransport.hpp>
#include <RCF/TcpEndpoint.hpp>

namespace RCF {

    // FilterAdapter

    class FilterAdapter : public RCF::IdentityFilter
    {
    public:
        FilterAdapter(TcpAsioServerTransport::SessionState &sessionState) :
            mSessionState(sessionState)
        {}

    private:
        void read(
            const ByteBuffer &byteBuffer,
            std::size_t bytesRequested)
        {
            mSessionState.read(byteBuffer, bytesRequested);
        }

        void write(
            const std::vector<ByteBuffer> &byteBuffers)
        {
            mSessionState.write(byteBuffers);
        }

        void onReadCompleted(
            const ByteBuffer &byteBuffer,
            int error)
        {
            boost::system::error_code ec(error, boost::system::get_system_category());
            mSessionState.onReadWrite(byteBuffer.getLength(), ec);
        }

        void onWriteCompleted(
            std::size_t bytesTransferred,
            int error)
        {
            boost::system::error_code ec(error, boost::system::get_system_category());
            mSessionState.onReadWrite(bytesTransferred, ec);
        }

        const FilterDescription &getFilterDescription() const
        {
            RCF_ASSERT(0);
            return * (const FilterDescription *) NULL;
        }

        TcpAsioServerTransport::SessionState &mSessionState;
    };

    // TcpAsioProactor

    TcpAsioServerTransport::TcpAsioProactor::TcpAsioProactor(TcpAsioServerTransport &tcpAsioServerTransport) :
        mTcpAsioServerTransport(tcpAsioServerTransport)
    {}

    void TcpAsioServerTransport::TcpAsioProactor::postRead()
    {
        SessionStatePtr sessionStatePtr( mSessionStateWeakPtr.lock() );
        RCF_ASSERT(sessionStatePtr.get());

        sessionStatePtr->mState = SessionState::ReadingDataCount;
        sessionStatePtr->getUniqueReadBuffer().resize(4);
        sessionStatePtr->mReadBufferRemaining = 4;
        sessionStatePtr->invokeAsyncRead();
    }

    ByteBuffer TcpAsioServerTransport::TcpAsioProactor::getReadByteBuffer()
    {
        SessionStatePtr sessionStatePtr( mSessionStateWeakPtr.lock() );
        RCF_ASSERT(sessionStatePtr.get());

        return ByteBuffer(
            &(*sessionStatePtr->mReadBufferPtr)[0],
            (*sessionStatePtr->mReadBufferPtr).size(),
            sessionStatePtr->mReadBufferPtr);
    }

    void TcpAsioServerTransport::TcpAsioProactor::postWrite(const std::vector<ByteBuffer> &byteBuffers)
    {
        SessionStatePtr sessionStatePtr( mSessionStateWeakPtr.lock() );
        RCF_ASSERT(sessionStatePtr.get());

        RCF_ASSERT(
            sessionStatePtr->mState == SessionState::Ready)
            (sessionStatePtr->mState);

        BOOST_STATIC_ASSERT(sizeof(unsigned int) == 4);

        std::vector<ByteBuffer> &writeByteBuffers =
            sessionStatePtr->mWriteByteBuffers;

        writeByteBuffers.resize(0);

        std::copy(
            byteBuffers.begin(),
            byteBuffers.end(),
            std::back_inserter(writeByteBuffers));

        int messageSize = static_cast<int>(RCF::lengthByteBuffers(byteBuffers));
        
        ByteBuffer &byteBuffer = writeByteBuffers.front();

        RCF_ASSERT(
            byteBuffer.getLeftMargin() >= 4)
            (byteBuffer.getLeftMargin());

        byteBuffer.expandIntoLeftMargin(4);
        memcpy(byteBuffer.getPtr(), &messageSize, 4);
        RCF::machineToNetworkOrder(byteBuffer.getPtr(), 4, 1);

        sessionStatePtr->mState = SessionState::WritingData;
        sessionStatePtr->mWriteBufferRemaining = RCF::lengthByteBuffers(writeByteBuffers);
        sessionStatePtr->invokeAsyncWrite();

    }

    void TcpAsioServerTransport::TcpAsioProactor::postClose()
    {
        SessionStatePtr sessionStatePtr( mSessionStateWeakPtr.lock() );
        RCF_ASSERT(sessionStatePtr.get());
        sessionStatePtr->close();
    }

    I_ServerTransport &TcpAsioServerTransport::TcpAsioProactor::getServerTransport()
    {
        return mTcpAsioServerTransport;
    }

    const I_RemoteAddress &TcpAsioServerTransport::TcpAsioProactor::getRemoteAddress()
    {
        SessionStatePtr sessionStatePtr( mSessionStateWeakPtr.lock() );
        RCF_ASSERT(sessionStatePtr.get());
        return sessionStatePtr->mIpAddress;
    }

    void TcpAsioServerTransport::TcpAsioProactor::setTransportFilters(
        const std::vector<FilterPtr> &filters)
    {
        SessionStatePtr sessionStatePtr( mSessionStateWeakPtr.lock() );
        RCF_ASSERT(sessionStatePtr.get());
        sessionStatePtr->setTransportFilters(filters);
    }

    void TcpAsioServerTransport::TcpAsioProactor::getTransportFilters(
        std::vector<FilterPtr> &filters)
    {
        mSessionStateWeakPtr.lock()->getTransportFilters(filters);
    }

    // SessionState

    TcpAsioServerTransport::SessionState::SessionState(
        TcpAsioServerTransport &transport,
        DemuxerPtr demuxerPtr) :
            mSocketPtr(new Socket(*demuxerPtr)),
            mState(Ready),
            mReadBufferRemaining(RCF_DEFAULT_INIT),
            mWriteBufferRemaining(RCF_DEFAULT_INIT),
            mReflecting(RCF_DEFAULT_INIT),
            mTransport(transport),
            mFilterAdapterPtr(new FilterAdapter(*this)),
            mHasBeenClosed(RCF_DEFAULT_INIT)
    {
    }

    boost::system::error_code closeAsioSocket(boost::asio::ip::tcp::socket &s)
    {
        boost::system::error_code error;
        s.close(error);
        return error;
    }

    TcpAsioServerTransport::SessionState::~SessionState()
    {
        RCF_DTOR_BEGIN

        // TODO: invoke accept if appropriate
        // TODO: need a proper acceptex strategy in the first place
        //RCF_ASSERT(mState != Accepting);

        mTransport.unregisterSession(mWeakThisPtr);

        // close reflecting session if appropriate
        if (mReflecting)
        {
            SessionStatePtr sessionStatePtr(mReflecteeWeakPtr.lock());
            if (sessionStatePtr)
            {
                sessionStatePtr->close();
            }
        }

        RCF_DTOR_END;
    }

    std::vector<char> &TcpAsioServerTransport::SessionState::getReadBuffer()
    {
        if (!mReadBufferPtr)
        {
            mReadBufferPtr.reset( new std::vector<char>() );
        }
        return *mReadBufferPtr;
    }

    std::vector<char> &TcpAsioServerTransport::SessionState::getUniqueReadBuffer()
    {
        if (!mReadBufferPtr || !mReadBufferPtr.unique())
        {
            mReadBufferPtr.reset( new std::vector<char>() );
        }
        return *mReadBufferPtr;
    }

    ByteBuffer TcpAsioServerTransport::SessionState::getReadByteBuffer() const
    {
        return ByteBuffer(
            &(*mReadBufferPtr)[0],
            (*mReadBufferPtr).size(),
            mReadBufferPtr);
    }

    std::vector<char> &TcpAsioServerTransport::SessionState::getReadBufferSecondary()
    {
        if (!mReadBufferSecondaryPtr)
        {
            mReadBufferSecondaryPtr.reset( new std::vector<char>() );
        }
        return *mReadBufferSecondaryPtr;
    }

    std::vector<char> &TcpAsioServerTransport::SessionState::getUniqueReadBufferSecondary()
    {
        if (!mReadBufferSecondaryPtr || !mReadBufferSecondaryPtr.unique())
        {
            mReadBufferSecondaryPtr.reset( new std::vector<char>() );
        }
        return *mReadBufferSecondaryPtr;
    }

    ByteBuffer TcpAsioServerTransport::SessionState::getReadByteBufferSecondary() const
    {
        return ByteBuffer(
            &(*mReadBufferSecondaryPtr)[0],
            (*mReadBufferSecondaryPtr).size(),
            mReadBufferSecondaryPtr);
    }

    void TcpAsioServerTransport::SessionState::read(
        const ByteBuffer &byteBuffer,
        std::size_t bytesRequested)
    {

        RCF2_TRACE("")(this);

        if (byteBuffer.getLength() == 0)
        {
            std::vector<char> &vec = getUniqueReadBufferSecondary();
            vec.resize(bytesRequested);
            mTempByteBuffer = getReadByteBufferSecondary();
        }
        else
        {
            mTempByteBuffer = ByteBuffer(byteBuffer, 0, bytesRequested);
        }

        RCF_ASSERT(
            bytesRequested <= mTempByteBuffer.getLength())
            (bytesRequested)(mTempByteBuffer.getLength());

        char *buffer = mTempByteBuffer.getPtr();
        std::size_t bufferLen = mTempByteBuffer.getLength();

        Lock lock(mMutex);
        if (!mHasBeenClosed)
        {
            mSocketPtr->async_receive(
                boost::asio::buffer(buffer, bufferLen),
                0,
                boost::bind(
                    &SessionState::onReadCompletion,
                    shared_from_this(),
                    boost::asio::placeholders::error,
                    boost::asio::placeholders::bytes_transferred));
        }
    }

    void TcpAsioServerTransport::SessionState::write(
        const std::vector<ByteBuffer> &byteBuffers)
    {
        RCF2_TRACE("")(this);

        // TODO: send all the buffers at once
        // ...

        RCF_ASSERT(!byteBuffers.empty());
        char *buffer = byteBuffers.front().getPtr();
        std::size_t bufferLen = byteBuffers.front().getLength();

        Lock lock(mMutex);
        if (!mHasBeenClosed)
        {
            boost::asio::async_write(
                *mSocketPtr,
                boost::asio::buffer(buffer, bufferLen),
                boost::bind(
                &SessionState::onWriteCompletion,
                shared_from_this(),
                boost::asio::placeholders::error,
                boost::asio::placeholders::bytes_transferred));
        }
    }

    // TODO: merge onReadCompletion/onWriteCompletion into one function

    void TcpAsioServerTransport::SessionState::onReadCompletion(const boost::system::error_code &error, size_t bytesTransferred)
    {
        RCF2_TRACE("")(this)(bytesTransferred);
        if (!error && !mTransport.mStopFlag)
        {
            if (mReflecting)
            {
                boost::system::error_code ec;
                onReflectedReadWrite(ec, bytesTransferred);
            }
            else
            {
                SetCurrentSessionGuard guard(mSessionPtr);
                ByteBuffer byteBuffer(mTempByteBuffer.release(), 0, bytesTransferred);

                boost::system::error_code ec;
                mTransportFilters.empty() ?
                    onReadWrite(bytesTransferred, ec) : // used to be boost::asio::error(error)
                    mTransportFilters.back()->onReadCompleted(byteBuffer, 0);
            }
        }
    }

    void TcpAsioServerTransport::SessionState::onWriteCompletion(const boost::system::error_code &error, size_t bytesTransferred)
    {
        RCF2_TRACE("")(this)(bytesTransferred);
        if (!error && !mTransport.mStopFlag)
        {
            if (mReflecting)
            {
                if (mReflecteePtr)
                {
                    mReflecteePtr.reset();
                }
                boost::system::error_code ec;
                onReflectedReadWrite(ec, bytesTransferred);
            }
            else
            {
                SetCurrentSessionGuard guard(mSessionPtr);
                boost::system::error_code ec;
                mTransportFilters.empty() ?
                    onReadWrite(bytesTransferred, ec) : // used to be boost::asio::error(error)
                    mTransportFilters.back()->onWriteCompleted(bytesTransferred, 0);
            }
        }
    }

    void TcpAsioServerTransport::SessionState::setTransportFilters(
        const std::vector<FilterPtr> &filters)
    {

        mTransportFilters.assign(filters.begin(), filters.end());
        connectFilters(mTransportFilters);
        if (!mTransportFilters.empty())
        {
            mTransportFilters.front()->setPreFilter( *mFilterAdapterPtr );
            mTransportFilters.back()->setPostFilter( *mFilterAdapterPtr );
        }
    }

    void TcpAsioServerTransport::SessionState::getTransportFilters(std::vector<FilterPtr> &filters)
    {
        filters = mTransportFilters;
    }

    void TcpAsioServerTransport::SessionState::invokeAsyncRead()
    {
        RCF2_TRACE("")(this);

        ByteBuffer byteBuffer(
            getReadByteBuffer(),
            getReadByteBuffer().getLength()-mReadBufferRemaining);

        mTransportFilters.empty() ?
            read(byteBuffer, mReadBufferRemaining) :
            mTransportFilters.front()->read(byteBuffer, mReadBufferRemaining);
    }

    void TcpAsioServerTransport::SessionState::invokeAsyncWrite()
    {
        RCF2_TRACE("")(this);

        // TODO: thread-local
        std::vector<RCF::ByteBuffer> byteBuffers;
        sliceByteBuffers(
            byteBuffers,
            mWriteByteBuffers,
            lengthByteBuffers(mWriteByteBuffers)-mWriteBufferRemaining);

        mTransportFilters.empty() ?
            write(byteBuffers) :
            mTransportFilters.front()->write(byteBuffers);

    }

    void TcpAsioServerTransport::SessionState::invokeAsyncAccept()
    {
        RCF2_TRACE("")(this);

        mState = SessionState::Accepting;

        mTransport.mAcceptorPtr->async_accept(
            *mSocketPtr,
            boost::bind(
                &SessionState::onAccept,
                shared_from_this(),
                boost::asio::placeholders::error));
    }

    void TcpAsioServerTransport::SessionState::onAccept(const boost::system::error_code& error)
    {
        RCF2_TRACE("")(this);

        if (mTransport.mStopFlag)
        {
            return;
        }

        if (!error)
        {
            // save the remote address in the SessionState object
            boost::asio::ip::tcp::endpoint endpoint = mSocketPtr->remote_endpoint();
            sockaddr_in addr;
            memset(&addr, 0, sizeof(addr));
            addr.sin_family = AF_INET;
            addr.sin_port = htons(endpoint.port());
            addr.sin_addr.s_addr = htonl(endpoint.address().to_v4().to_ulong());
            mIpAddress = IpAddress(addr);
            mState = SessionState::WritingData;

            // create a new SessionState, and do an accept on that
            mTransport.createSessionState()->invokeAsyncAccept();

            // set current RCF session
            SetCurrentSessionGuard guard(mSessionPtr);

            if (mTransport.isClientAddrAllowed(addr))
            {
                // Check the connection limit.
                bool allowConnect = true;
                std::size_t connectionLimit = mTransport.getConnectionLimit();
                if (connectionLimit)
                {
                    Lock lock(mTransport.mSessionsMutex);
                    RCF_ASSERT(mTransport.mSessions.size() <= 1 + 1 + connectionLimit);
                    if (mTransport.mSessions.size() == 1 + 1 + connectionLimit)
                    {
                        allowConnect = false;
                    }
                }

                if (allowConnect)
                {
                    // start things rolling by faking a completed write operation
                    onReadWrite(0, boost::system::error_code());
                }
                else
                {
                    sendServerError(RcfError_ConnectionLimitExceeded);
                }
            }
        }
        else if (
            error == boost::asio::error::connection_aborted ||
            error == boost::asio::error::operation_aborted)
        {
            invokeAsyncAccept();
        }
    }

    void onError(boost::system::error_code &error1, const boost::system::error_code &error2)
    {
        error1 = error2;
    }

    void TcpAsioServerTransport::SessionState::sendServerError(int error)
    {
        // TODO: improve this a bit

        ByteBuffer byteBuffer = encodeServerError(error);
        byteBuffer.expandIntoLeftMargin(4);

        boost::uint32_t len = 
            static_cast<boost::uint32_t>(byteBuffer.getLength()-4);
        memcpy(byteBuffer.getPtr(), &len, 4);
        machineToNetworkOrder(byteBuffer.getPtr(), 4, 1);

        char *buffer = byteBuffer.getPtr();
        std::size_t bufferLen = byteBuffer.getLength();

        boost::system::error_code err;
        boost::asio::write(
            *mSocketPtr,
            boost::asio::buffer(buffer, bufferLen),
            boost::asio::transfer_all(),
            err);

        mSocketPtr->shutdown(boost::asio::socket_base::shutdown_send);
    }

    void TcpAsioServerTransport::SessionState::onReadWrite(size_t bytesTransferred, const boost::system::error_code& error)
    {
        RCF_ASSERT(!error);
        RCF_ASSERT(!mReflecting);
        {
            switch(mState)
            {
            case SessionState::ReadingDataCount:
            case SessionState::ReadingData:

                RCF_ASSERT(
                    bytesTransferred <= mReadBufferRemaining)
                    (bytesTransferred)(mReadBufferRemaining);

                mReadBufferRemaining -= bytesTransferred;
                if (mReadBufferRemaining > 0)
                {
                    invokeAsyncRead();
                }
                else
                {
                    RCF_ASSERT(mReadBufferRemaining == 0)(mReadBufferRemaining);
                    if (mState == SessionState::ReadingDataCount)
                    {
                        std::vector<char> &readBuffer = getReadBuffer();
                        RCF_ASSERT(readBuffer.size() == 4)(readBuffer.size());

                        unsigned int packetLength = 0;
                        memcpy(&packetLength, &readBuffer[0], 4);
                        networkToMachineOrder(&packetLength, 4, 1);
                        
                        if (packetLength <= mTransport.getMaxMessageLength())
                        {
                            readBuffer.resize(packetLength);
                            mReadBufferRemaining = packetLength;
                            mState = SessionState::ReadingData;
                            invokeAsyncRead();
                        }
                        else
                        {
                            sendServerError(RcfError_ServerMessageLength);
                        }

                    }
                    else if (mState == SessionState::ReadingData)
                    {
                        mState = SessionState::Ready;
                        mTransport.getSessionManager().onReadCompleted(getSessionPtr());

                        if (mTransport.mInterrupt)
                        {
                            mTransport.mInterrupt = false;
                            mTransport.mDemuxerPtr->stop();
                        }
                    }
                }
                break;


            case SessionState::WritingData:

                RCF_ASSERT(
                    bytesTransferred <= mWriteBufferRemaining)
                    (bytesTransferred)(mWriteBufferRemaining);

                mWriteBufferRemaining -= bytesTransferred;
                if (mWriteBufferRemaining > 0)
                {
                    invokeAsyncWrite();
                }
                else
                {
                    mState = SessionState::Ready;
                    mTransport.getSessionManager().onWriteCompleted(getSessionPtr());

                    if (mTransport.mInterrupt)
                    {
                        mTransport.mInterrupt = false;
                        mTransport.mDemuxerPtr->stop();
                    }
                }
                break;

            default:
                RCF_ASSERT(0);
            }
        }
    }

    void TcpAsioServerTransport::SessionState::onReflectedReadWrite(
        const boost::system::error_code& error,
        size_t bytesTransferred)
    {
        RCF2_TRACE("")(this);
        RCF_UNUSED_VARIABLE(error);

        RCF_ASSERT(
            mState == SessionState::ReadingData ||
            mState == SessionState::ReadingDataCount ||
            mState == SessionState::WritingData)
            (mState);

        // TODO: whether sync is needed, depends on boost::asio implementation
        RCF_ASSERT(mReflecting);

        if (mState == SessionState::WritingData)
        {
            mState = SessionState::ReadingData;
            std::vector<char> &readBuffer = getReadBuffer();
            readBuffer.resize(8*1024);

            char *buffer = &readBuffer[0];
            std::size_t bufferLen = readBuffer.size();

            Lock lock(mMutex);
            if (!mHasBeenClosed)
            {
                mSocketPtr->async_receive(
                    boost::asio::buffer(buffer, bufferLen),
                    0,
                    boost::bind(
                        &SessionState::onReadCompletion,
                        shared_from_this(),
                        boost::asio::placeholders::error,
                        boost::asio::placeholders::bytes_transferred));
            }
        }
        else if (
            mState == SessionState::ReadingData ||
            mState == SessionState::ReadingDataCount)
        {
            mState = SessionState::WritingData;
            std::vector<char> &readBuffer = getReadBuffer();

            char *buffer = &readBuffer[0];
            std::size_t bufferLen = bytesTransferred;

            // mReflecteePtr will be nulled in onWriteCompletion()
            // (otherwise we could easily end up with a cycle)
            RCF_ASSERT(!mReflecteePtr);
            mReflecteePtr = mReflecteeWeakPtr.lock();
            if (mReflecteePtr)
            {
                RCF_ASSERT(mReflecteePtr->mReflecting);

                Lock lock(mReflecteePtr->mMutex);
                if (!mReflecteePtr->mHasBeenClosed)
                {
                    // TODO: if this can throw, then we need a scope_guard
                    // to reset mReflecteePtr
                    boost::asio::async_write(
                        *(mReflecteePtr->mSocketPtr),
                        boost::asio::buffer(buffer, bufferLen),
                        boost::bind(
                            &SessionState::onWriteCompletion,
                            shared_from_this(),
                            boost::asio::placeholders::error,
                            boost::asio::placeholders::bytes_transferred));
                }
            }
        }
    }

    // TcpAsioServerTransport

    TcpAsioServerTransport::SessionStatePtr TcpAsioServerTransport::createSessionState()
    {
        RCF2_TRACE("");

        SessionStatePtr sessionStatePtr( new SessionState(*this, mDemuxerPtr) );
        SessionPtr sessionPtr( getSessionManager().createSession() );
        TcpAsioProactorPtr tcpAsioProactorPtr( new TcpAsioProactor(*this) );
        sessionPtr->setProactorPtr( ProactorPtr(tcpAsioProactorPtr) );
        tcpAsioProactorPtr->mSessionStateWeakPtr = sessionStatePtr;
        sessionStatePtr->setSessionPtr(sessionPtr);
        sessionStatePtr->mWeakThisPtr = sessionStatePtr;
        registerSession(sessionStatePtr->mWeakThisPtr);
        return sessionStatePtr;
    }

    int TcpAsioServerTransport::getPort() const
    {
        return mPort;
    }

    // I_ServerTransportEx implementation

    ClientTransportAutoPtr TcpAsioServerTransport::createClientTransport(const I_Endpoint &endpoint)
    {
        RCF2_TRACE("");

        const TcpEndpoint &tcpEndpoint = dynamic_cast<const TcpEndpoint &>(endpoint);
        ClientTransportAutoPtr clientTransportAutoPtr(
            new TcpClientTransport(tcpEndpoint.getIp(), tcpEndpoint.getPort()));
        return clientTransportAutoPtr;
    }

    SessionPtr TcpAsioServerTransport::createServerSession(
        ClientTransportAutoPtr clientTransportAutoPtr,
        StubEntryPtr stubEntryPtr)
    {
        RCF2_TRACE("");

        TcpClientTransport *pTcpClientTransport =
            dynamic_cast<TcpClientTransport *>(clientTransportAutoPtr.get());

        if (pTcpClientTransport == NULL)
        {
            RCF_THROW(
                Exception("incompatible client transport"))
                (pTcpClientTransport)(typeid(*clientTransportAutoPtr));
        }

        TcpClientTransport &tcpClientTransport = *pTcpClientTransport;
        SessionStatePtr sessionStatePtr(createSessionState());
        SessionPtr sessionPtr(sessionStatePtr->getSessionPtr());

        // TODO: exception safety
        sessionStatePtr->mSocketPtr->assign(
            boost::asio::ip::tcp::v4(),
            tcpClientTransport.releaseFd());

        if (stubEntryPtr)
        {
            RcfSessionPtr rcfSessionPtr = 
                boost::static_pointer_cast<RcfSession>( 
                sessionStatePtr->mSessionPtr );

            rcfSessionPtr->setDefaultStubEntryPtr(stubEntryPtr);
        }

        sessionStatePtr->mState = SessionState::WritingData;
        sessionStatePtr->onReadWrite(0, boost::system::error_code());
        return sessionPtr;
    }

    void closeSocket(TcpAsioServerTransport::SocketPtr socketPtr)
    {
        socketPtr->close();
    }

    ClientTransportAutoPtr TcpAsioServerTransport::createClientTransport(SessionPtr sessionPtr)
    {
        RCF2_TRACE("");

        ProactorPtr proactorPtr(sessionPtr->getProactorPtr());
        TcpAsioProactorPtr tcpAsioProactorPtr(boost::dynamic_pointer_cast<TcpAsioProactor>(proactorPtr));
        SessionStatePtr sessionStatePtr(tcpAsioProactorPtr->mSessionStateWeakPtr.lock());
        sessionStatePtr->mDemuxerPtr = mDemuxerPtr;

        int fd = sessionStatePtr->mSocketPtr->native();
        std::auto_ptr<TcpClientTransport> tcpClientTransport(
            new TcpClientTransport(fd));

        boost::asio::ip::tcp::endpoint endpoint = sessionStatePtr->mSocketPtr->remote_endpoint();
        sockaddr_in addr;
        memset(&addr, 0, sizeof(addr));
        addr.sin_family = AF_INET;
        addr.sin_port = htons(endpoint.port());
        addr.sin_addr.s_addr = htonl(endpoint.address().to_v4().to_ulong());
        tcpClientTransport->setRemoteAddr(addr);

        tcpClientTransport->setNotifyCloseFunctor( boost::bind(
            &TcpAsioServerTransport::notifyClose,
            this,
            SessionStateWeakPtr(sessionStatePtr)));

        tcpClientTransport->setCloseFunctor( boost::bind(
            closeSocket,
            sessionStatePtr->mSocketPtr));

        return ClientTransportAutoPtr(tcpClientTransport.release());
    }

    bool TcpAsioServerTransport::reflect(const SessionPtr &sessionPtr1, const SessionPtr &sessionPtr2)
    {
        RCF2_TRACE("");

        ProactorPtr proactorPtr1(sessionPtr1->getProactorPtr());
        TcpAsioProactorPtr tcpAsioProactorPtr1(boost::dynamic_pointer_cast<TcpAsioProactor>(proactorPtr1));
        SessionStatePtr sessionStatePtr1(tcpAsioProactorPtr1->mSessionStateWeakPtr.lock());

        ProactorPtr proactorPtr2(sessionPtr2->getProactorPtr());
        TcpAsioProactorPtr tcpAsioProactorPtr2(boost::dynamic_pointer_cast<TcpAsioProactor>(proactorPtr2));
        SessionStatePtr sessionStatePtr2(tcpAsioProactorPtr2->mSessionStateWeakPtr.lock());

        sessionStatePtr1->mReflecteeWeakPtr = sessionStatePtr2;
        sessionStatePtr2->mReflecteeWeakPtr = sessionStatePtr1;

        sessionStatePtr1->mReflecting = true;
        sessionStatePtr2->mReflecting = true;


        return true;
    }

    bool TcpAsioServerTransport::isConnected(const SessionPtr &sessionPtr)
    {
        ProactorPtr proactorPtr(sessionPtr->getProactorPtr());
        TcpAsioProactorPtr tcpAsioProactorPtr( boost::dynamic_pointer_cast<TcpAsioProactor>(proactorPtr));
        SessionStatePtr sessionStatePtr( tcpAsioProactorPtr->mSessionStateWeakPtr.lock() );
        return sessionStatePtr.get() && isFdConnected(sessionStatePtr->mSocketPtr->native());
    }

    // I_Service implementation

    void TcpAsioServerTransport::open()
    {
        RCF2_TRACE("");

        mInterrupt = false;
        mStopFlag = false;
        mDemuxerPtr.reset( new Demuxer );
        mCycleTimerPtr.reset( new DeadlineTimer(*mDemuxerPtr) );

        if (mPort >= 0)
        {

            //mAcceptorPtr = SocketAcceptorPtr( new SocketAcceptor(
            //    *mDemuxerPtr,
            //    boost::asio::ip::tcp::endpoint(
            //        boost::asio::ip::address::from_string(getNetworkInterface()),
            //        static_cast<unsigned short>(mPort))));

            boost::asio::ip::tcp::endpoint endpoint(
                boost::asio::ip::address::from_string(getNetworkInterface()),
                static_cast<unsigned short>(mPort));

            mAcceptorPtr = SocketAcceptorPtr( new SocketAcceptor(*mDemuxerPtr));
            mAcceptorPtr->open(endpoint.protocol());
            mAcceptorPtr->set_option(boost::asio::socket_base::reuse_address(true));
            mAcceptorPtr->bind(endpoint);
            mAcceptorPtr->listen();

            // retrieve the port number, if it's generated by the system
            int fd = mAcceptorPtr->native();
            std::string networkInterface = getNetworkInterface();
            if (mPort == 0 && (mPort = getFdPort(fd)) == 0)
            {
                int err = Platform::OS::BsdSockets::GetLastError();
                RCF_THROW(Exception(
                    RcfError_Socket, err, RcfSubsystem_Os, "getsockname() failed"))
                    (fd)(mPort)(networkInterface);
            }

        }
    }

    void TcpAsioServerTransport::SessionState::setSessionPtr(SessionPtr sessionPtr)    
    { 
        mSessionPtr = sessionPtr; 
    }
    
    SessionPtr TcpAsioServerTransport::SessionState::getSessionPtr()
    { 
        return mSessionPtr; 
    }

    TcpAsioServerTransport::SocketPtr TcpAsioServerTransport::SessionState::getSocketPtr() 
    { 
        return mSocketPtr; 
    }

    void TcpAsioServerTransport::SessionState::close()
    {
        RCF2_TRACE("")(this);

        Lock lock(mMutex);
        if (!mHasBeenClosed)
        {
            mSocketPtr->close();
            mDemuxerPtr.reset();
            mHasBeenClosed = true;
        }
    }

    void TcpAsioServerTransport::notifyClose(
        SessionStateWeakPtr sessionStateWeakPtr)
    {
        SessionStatePtr sessionStatePtr(sessionStateWeakPtr.lock());
        if (sessionStatePtr)
        {
            Lock lock(sessionStatePtr->mMutex);
            sessionStatePtr->mHasBeenClosed = true;
        }
    }

    void TcpAsioServerTransport::close()
    {
        RCF2_TRACE("");

        mAcceptorPtr.reset();
        mCycleTimerPtr.reset();

        mStopFlag = true;
        cancelOutstandingIo();

        mDemuxerPtr->reset();
        std::size_t items = mDemuxerPtr->run();
        while (items)
        {
            mDemuxerPtr->reset();
            items = mDemuxerPtr->run();
        }

        mDemuxerPtr.reset();
    }

    bool TcpAsioServerTransport::cycle(
        int timeoutMs,
        const volatile bool &stopFlag,
        bool returnEarly)
    {
        RCF2_TRACE("Entering cycle()");

        RCF_UNUSED_VARIABLE(stopFlag);
        RCF_ASSERT(timeoutMs >= -1)(timeoutMs);

        mInterrupt = returnEarly;
        if (timeoutMs != -1)
        {
            mCycleTimerPtr->cancel();
            mCycleTimerPtr->expires_from_now(boost::posix_time::milliseconds(timeoutMs)); // replacing timeoutMs with 1000 makes Test_Minimal work with cw...
            mCycleTimerPtr->async_wait( boost::bind(&TcpAsioServerTransport::stopCycle, this, boost::asio::placeholders::error));
        }

        mDemuxerPtr->reset();
        mDemuxerPtr->run();

        RCF2_TRACE("Exiting cycle()");

        return false;
    }

    void TcpAsioServerTransport::stopCycle(const boost::system::error_code &error)
    {
        RCF2_TRACE("");

        if (!error)
        {
            mDemuxerPtr->stop();
        }
    }

    void TcpAsioServerTransport::stop()
    {
        mDemuxerPtr->stop();
    }

    void TcpAsioServerTransport::onServiceAdded(RcfServer &server)
    {
        setServer(server);
        WriteLock writeLock( getTaskEntriesMutex() );
        getTaskEntries().clear();
        getTaskEntries().push_back(
            TaskEntry(
            boost::bind(&TcpAsioServerTransport::cycle, this, _1, _2, _3),
            boost::bind(&TcpAsioServerTransport::stop, this),
            "RCF Asio server"));
    }

    void TcpAsioServerTransport::onServiceRemoved(RcfServer &)
    {}

    void TcpAsioServerTransport::onServerOpen(RcfServer &)
    {
        open();
    }

    void TcpAsioServerTransport::onServerClose(RcfServer &)
    {
        close();
    }

    void TcpAsioServerTransport::onServerStart(RcfServer &)
    {
        mStopFlag = false;
        if (mAcceptorPtr)
        {
            createSessionState()->invokeAsyncAccept();
        }
    }

    void TcpAsioServerTransport::onServerStop(RcfServer &)
    {
    }

    void TcpAsioServerTransport::setServer(RcfServer &server)
    {
        pServer = &server;
    }

    RcfServer &TcpAsioServerTransport::getServer()
    {
        return *pServer;
    }

    I_SessionManager &TcpAsioServerTransport::getSessionManager()
    {
        return *pServer;
    }

    TcpAsioServerTransport::TcpAsioServerTransport(int port) :
        mDemuxerPtr(),
        mCycleTimerPtr(),
        mInterrupt(RCF_DEFAULT_INIT),
        mReadWriteMutexPtr( new ReadWriteMutex(ReaderPriority) ),
        mPort(port),
        mAcceptorPtr(),
        mStopFlag(RCF_DEFAULT_INIT),
        pServer(RCF_DEFAULT_INIT)
    {}

    TcpAsioServerTransport::TcpAsioServerTransport(const std::string &networkInterface, int port) :
        mDemuxerPtr(),
        mCycleTimerPtr(),
        mInterrupt(RCF_DEFAULT_INIT),
        mReadWriteMutexPtr( new ReadWriteMutex(ReaderPriority) ),
        mPort(port),
        mAcceptorPtr(),
        mStopFlag(RCF_DEFAULT_INIT),
        pServer(RCF_DEFAULT_INIT)
    {
        setNetworkInterface(networkInterface);
    }

    ServerTransportPtr TcpAsioServerTransport::clone()
    {
        return ServerTransportPtr(new TcpAsioServerTransport(mPort));
    }

    void TcpAsioServerTransport::registerSession(SessionStateWeakPtr session)
    {
        Lock lock(mSessionsMutex);
        mSessions.insert(session);
    }

    void TcpAsioServerTransport::unregisterSession(SessionStateWeakPtr session)
    {
        Lock lock(mSessionsMutex);
        mSessions.erase(session);
    }

    void TcpAsioServerTransport::cancelOutstandingIo()
    {
        Lock lock(mSessionsMutex);
        std::for_each( 
            mSessions.begin(), 
            mSessions.end(), 
            boost::bind(&TcpAsioServerTransport::closeSessionState, this, _1));
    }

    void TcpAsioServerTransport::closeSessionState(SessionStateWeakPtr sessionStateWeakPtr)
    {
        SessionStatePtr sessionStatePtr(sessionStateWeakPtr.lock());
        if (sessionStatePtr)
        {
            sessionStatePtr->close();
        }
    }

} // namespace RCF
