
//******************************************************************************
// RCF - Remote Call Framework
// Copyright (c) 2005 - 2008, Jarl Lindrud. All rights reserved.
// Consult your license for conditions of use.
// Version: 0.9d
// Contact: jarl.lindrud <at> gmail.com 
//******************************************************************************

#include <RCF/Win32NamedPipeClientTransport.hpp>

#include <RCF/ByteOrdering.hpp>

namespace RCF {

    Win32NamedPipeClientTransport::Win32NamedPipeClientTransport(
        const Win32NamedPipeClientTransport & rhs) :
            I_ClientTransport(rhs),
            mPipeName(rhs.mPipeName),
            mhPipe(INVALID_HANDLE_VALUE),
            mhEvent(INVALID_HANDLE_VALUE)
    {
        HANDLE hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
        DWORD dwErr = GetLastError();
        RCF_VERIFY( hEvent, Exception(RcfError_Pipe, dwErr));

        mhEvent = hEvent;
    }

    Win32NamedPipeClientTransport::Win32NamedPipeClientTransport(
        const tstring & pipeName,
        const tstring & machine) :
            mhPipe(INVALID_HANDLE_VALUE),
            mhEvent(INVALID_HANDLE_VALUE)
    {
        if (pipeName.at(0) == _T('\\'))
        {
            mPipeName = pipeName;
        }
        else if (machine.empty())
        {
            mPipeName = _T("\\\\.\\pipe\\") + pipeName;
        }
        else
        {
            mPipeName = _T("\\\\") + machine + _T("\\pipe\\") + pipeName;
        }

        HANDLE hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
        DWORD dwErr = GetLastError();
        RCF_VERIFY( hEvent, Exception(RcfError_Pipe, dwErr));

        mhEvent = hEvent;
    }

    Win32NamedPipeClientTransport::~Win32NamedPipeClientTransport()
    {
        RCF_DTOR_BEGIN
            disconnect();

            if (mhEvent != INVALID_HANDLE_VALUE)
            {
                BOOL ok = CloseHandle(mhEvent);
                DWORD dwErr = GetLastError();
                RCF_VERIFY(ok, Exception(RcfError_Pipe, dwErr));
            }
        RCF_DTOR_END
    }

    ClientTransportAutoPtr Win32NamedPipeClientTransport::clone() const
    {
        return ClientTransportAutoPtr( 
            new Win32NamedPipeClientTransport(*this));
    }

    EndpointPtr Win32NamedPipeClientTransport::getEndpointPtr() const
    {
        // Not applicable
        return EndpointPtr();
    }

    int Win32NamedPipeClientTransport::send(
        const std::vector<ByteBuffer> &data, 
        unsigned int timeoutMs)
    {
        RCF_VERIFY(
            isConnected(),
            RCF::Exception(RcfError_NotConnected));

        // Prepend the message length
        ByteBuffer &buffer = const_cast<ByteBuffer &>(data.front());

        unsigned int messageLength = static_cast<unsigned int>(
            lengthByteBuffers(data));
        
        BOOST_STATIC_ASSERT(sizeof(unsigned int) == 4);
        RCF_ASSERT(buffer.getLeftMargin() >= 4)(buffer.getLeftMargin());
        buffer.expandIntoLeftMargin(4);
        memcpy(buffer.getPtr(), &messageLength, 4);
        RCF::machineToNetworkOrder(buffer.getPtr(), 4, 1);

        unsigned int endTimeMs = getCurrentTimeMs() + timeoutMs;

        // Send it
        for (std::size_t i=0; i<data.size(); ++i)
        {
            const ByteBuffer &buffer = data[i];
            std::size_t bytesWritten = 0;
            std::size_t bytesTotal = buffer.getLength();

            while (bytesWritten < bytesTotal)
            {
                RCF_VERIFY(
                    generateTimeoutMs(endTimeMs),
                    Exception(RcfError_ClientWriteTimeout))
                    (timeoutMs);

                DWORD count = 0;

                BOOL ok = WriteFile( 
                    mhPipe,
                    buffer.getPtr(),
                    static_cast<DWORD>(buffer.getLength()),
                    &count,
                    NULL);

                DWORD dwErr = GetLastError();

                RCF_VERIFY(
                    ok,
                    Exception(
                        RcfError_ClientWriteFail,
                        dwErr,
                        RcfSubsystem_Os,
                        "WriteFile() failed"))
                    (mhPipe);

                bytesWritten += count;
            }

            RCF_ASSERT( bytesWritten == bytesTotal );

        }

        return 1;
    }

    void Win32NamedPipeClientTransport::timedReceive(
        char *buffer, 
        std::size_t len, 
        unsigned int endTimeMs)
    {
        DWORD dwLen = static_cast<DWORD>(len);

        DWORD dwPos = 0;
        while (dwPos < dwLen && generateTimeoutMs(endTimeMs))
        {
            BOOL ok = ResetEvent(mhEvent);
            DWORD dwErr = GetLastError();
            RCF_VERIFY(ok, Exception(RcfError_Pipe, dwErr));

            OVERLAPPED overlapped = {0};
            overlapped.hEvent = mhEvent;

            DWORD dwRead = 0;
            ok = ReadFile(mhPipe, &buffer[dwPos], dwLen-dwPos, &dwRead, &overlapped);
            dwErr = GetLastError();

            if (!ok)
            {
                RCF_VERIFY( 
                    dwErr == ERROR_IO_PENDING ||
                    dwErr == WSA_IO_PENDING ||
                    dwErr == ERROR_MORE_DATA,
                    Exception(RcfError_ClientReadFail));
            }

            DWORD dwRet = WAIT_TIMEOUT;
            while (dwRet == WAIT_TIMEOUT)
            {
                dwRet = WaitForSingleObject(overlapped.hEvent, 500);
                dwErr = GetLastError();

                RCF_VERIFY( 
                    dwRet == WAIT_OBJECT_0 || dwRet == WAIT_TIMEOUT, 
                    Exception(RcfError_Pipe, dwErr));
            }
            RCF_ASSERT(dwRet == WAIT_OBJECT_0);

            dwRead = 0;
            ok = GetOverlappedResult(mhPipe, &overlapped, &dwRead, FALSE);
            dwErr = GetLastError();
            RCF_VERIFY(ok && dwRead > 0, Exception(RcfError_Pipe, dwErr));

            dwPos += dwRead;
            RCF_ASSERT(dwPos <= dwLen);
        }

        RCF_VERIFY(
            dwPos == dwLen,
            Exception(RcfError_ClientReadTimeout))
            (endTimeMs)(dwPos)(dwLen);

    }

    int Win32NamedPipeClientTransport::receive(
        ByteBuffer &byteBuffer, 
        unsigned int timeoutMs)
    {
        unsigned int endTimeMs = getCurrentTimeMs() + timeoutMs;

        unsigned int msgLength = 0;
        timedReceive( (char *) &msgLength, 4, endTimeMs);

        RCF::networkToMachineOrder(&msgLength, 4, 1);

        RCF_VERIFY(
            0 <= msgLength && msgLength <= getMaxMessageLength(),
            Exception(RcfError_ClientMessageLength))
            (msgLength)(getMaxMessageLength());

        if (!mReadBufferPtr || !mReadBufferPtr.unique())
        {
            mReadBufferPtr.reset(new std::vector<char>());
        }
        mReadBufferPtr->resize(msgLength);

        timedReceive(&(*mReadBufferPtr)[0], mReadBufferPtr->size(), endTimeMs);
        byteBuffer = ByteBuffer(mReadBufferPtr);
        return 1;
    }

    bool Win32NamedPipeClientTransport::isConnected()
    {
        bool connected = 
            mhPipe != INVALID_HANDLE_VALUE &&
            PeekNamedPipe(mhPipe, NULL,0, NULL, NULL, NULL) == TRUE;

        if (mhPipe != INVALID_HANDLE_VALUE && !connected)
        {
            // Will probably be ERROR_PIPE_NOT_CONNECTED.
            DWORD dwErr = GetLastError();
            RCF_UNUSED_VARIABLE(dwErr);
        }

        return connected;
    }

    void Win32NamedPipeClientTransport::connect(unsigned int timeoutMs)
    {
        if (!isConnected())
        {
            unsigned int endTimeMs = getCurrentTimeMs() + timeoutMs;
            HANDLE hPipe = INVALID_HANDLE_VALUE;
            while (hPipe == INVALID_HANDLE_VALUE && generateTimeoutMs(endTimeMs)) 
            { 
                hPipe = CreateFile( 
                    mPipeName.c_str(),   // pipe name 
                    GENERIC_READ |  // read and write access 
                    GENERIC_WRITE, 
                    0,              // no sharing 
                    NULL,           // default security attributes
                    OPEN_EXISTING,  // opens existing pipe 
                    0,              // default attributes 
                    NULL);          // no template file 

                DWORD dwErr = GetLastError();

                if (hPipe == INVALID_HANDLE_VALUE && dwErr == ERROR_PIPE_BUSY) 
                {
                    DWORD timeoutMs = 100;
                    BOOL ok = WaitNamedPipe(mPipeName.c_str(), timeoutMs);
                    DWORD dwErr = GetLastError();
                    RCF_UNUSED_VARIABLE(ok);
                    RCF_UNUSED_VARIABLE(dwErr);
                }
                else if (hPipe == INVALID_HANDLE_VALUE)
                {
                    RCF_THROW(
                        Exception(
                            RcfError_ClientConnectFail,
                            dwErr,
                            RcfSubsystem_Os,
                            "CreateFile() failed"));
                }
            }
            if (hPipe == INVALID_HANDLE_VALUE)
            {
                RCF_THROW(Exception(RcfError_ClientConnectTimeout))(timeoutMs);
            }
            else
            {
                mhPipe = hPipe;
            }
        }
    }

    void Win32NamedPipeClientTransport::disconnect(unsigned int timeoutMs)
    {
        RCF_UNUSED_VARIABLE(timeoutMs);

        if (mhPipe != INVALID_HANDLE_VALUE)
        {
            BOOL ok = CloseHandle(mhPipe);
            DWORD dwErr = GetLastError();

            mhPipe = INVALID_HANDLE_VALUE;

            RCF_VERIFY(
                ok,
                Exception(
                    RcfError_Pipe,
                    dwErr,
                    RcfSubsystem_Os,
                    "CloseHandle() failed"))
                (mhPipe);
        }            
    }

    void Win32NamedPipeClientTransport::setTransportFilters(
        const std::vector<FilterPtr> &filters)
    {
        RCF_UNUSED_VARIABLE(filters);

        // Not implemented yet
    }

    void Win32NamedPipeClientTransport::getTransportFilters(
        std::vector<FilterPtr> &filters)
    {
        RCF_UNUSED_VARIABLE(filters);

        // Not implemented yet
    }

} // namespace RCF
