
// uncomment to enable VLD leak detection - will automatically link to required libs
//#include "vld.h"
//#include "vldapi.h"

#include <string>

#include <boost/bind.hpp>

#include <RCF/test/TestMinimal.hpp>

#include <RCF/Idl.hpp>
#include <RCF/RcfServer.hpp>
#include <RCF/Version.hpp>
#include <RCF/test/EndpointFactories.hpp>
#include <RCF/test/TransportFactories.hpp>
#include <RCF/util/CommandLine.hpp>

namespace Test_Versioning {

    enum MyEnum { One, Two, Three };

    class Echo
    {
    public:
        std::string echo(const std::string &s)
        {
            return s;
        }

        std::string echo(const std::string &s, int version)
        {
            int sessionVersion = RCF::getCurrentRcfSessionPtr()->getRcfRuntimeVersion();
            BOOST_CHECK( sessionVersion == version);
            return s;
        }

        MyEnum echo(MyEnum myEnum)
        {
            return myEnum;
        }

        int echo(int n)
        {
            return n;
        }
    };

    RCF_BEGIN(I_Echo, "I_Echo")
        RCF_METHOD_R2(std::string, echo, const std::string &, int)
    RCF_END(I_Echo)

    RCF_BEGIN(I_EchoWithInt, "I_Echo")
        RCF_METHOD_R2(std::string, echo, const std::string &, int)
        RCF_METHOD_R1(int, echo, int)
    RCF_END(I_EchoWithInt)

    RCF_BEGIN(I_EchoWithEnum, "I_Echo")
        RCF_METHOD_R2(std::string, echo, const std::string &, int)
        RCF_METHOD_R1(MyEnum, echo, MyEnum)
    RCF_END(I_EchoWithEnum)

} // namespace Test_Minimal

BOOST_BROKEN_COMPILER_TYPE_TRAITS_SPECIALIZATION(Test_Versioning::MyEnum)

int test_main(int argc, char **argv)
{

    printTestHeader(__FILE__);

    RCF::RcfInitDeinit rcfInitDeinit;

    using namespace Test_Versioning;

    util::CommandLineOption<std::string>    clIp("ip", util::PortNumbers::getSingleton().getIp(), "ip");
    util::CommandLineOption<int>            clPort("port", util::PortNumbers::getSingleton().getCurrent(), "port");
    util::CommandLine::getSingleton().parse(argc, argv);

    util::PortNumbers::getSingleton().setCurrent(clPort);
    util::PortNumbers::getSingleton().setIp(clIp);

    RCF::TcpEndpoint tcpEndpoint( util::PortNumbers::getSingleton().getNext());

    std::string s0 = "something special";

    Echo echo;
    RCF::RcfServer server(tcpEndpoint);
    server.bind( (I_Echo*) 0, echo);
    server.bind( (I_EchoWithInt*) 0, echo);
    server.bind( (I_EchoWithEnum*) 0, echo);
    server.start();

    RcfClient<I_Echo> client(tcpEndpoint);

    int currentVersion = RCF::getRuntimeVersion();
    BOOST_CHECK( client.echo(s0, currentVersion) == s0);
    BOOST_CHECK( client.getClientStub().getRcfRuntimeVersion() == currentVersion);

    for (int oldVersion = 1; oldVersion <= currentVersion; ++oldVersion)
    {
        // 1. client old, server new
        client.getClientStub().setRcfRuntimeVersion(oldVersion);
        server.setRcfRuntimeVersion(currentVersion);
        BOOST_CHECK( client.echo(s0, oldVersion) == s0);
        BOOST_CHECK( client.getClientStub().getRcfRuntimeVersion() == oldVersion);

        // 2. client new, server old
        client.getClientStub().setRcfRuntimeVersion(currentVersion);
        server.setRcfRuntimeVersion(oldVersion);
        BOOST_CHECK( client.getClientStub().getRcfRuntimeVersion() == currentVersion);
        BOOST_CHECK( client.echo(s0, oldVersion) == s0);
        BOOST_CHECK( client.getClientStub().getRcfRuntimeVersion() == oldVersion);

        client.getClientStub().setAutoVersioning(false);
        client.getClientStub().setRcfRuntimeVersion(currentVersion);
        try
        {
            client.echo(s0, currentVersion);
            BOOST_CHECK(oldVersion == currentVersion );
        }
        catch(const RCF::VersioningException &e)
        {
            BOOST_CHECK(oldVersion != currentVersion);
            BOOST_CHECK(e.getVersion() == oldVersion);
            BOOST_CHECK( client.getClientStub().getRcfRuntimeVersion() == currentVersion);
            client.getClientStub().setRcfRuntimeVersion(oldVersion);
            BOOST_CHECK( client.echo(s0, oldVersion) == s0);
            BOOST_CHECK( client.getClientStub().getRcfRuntimeVersion() == oldVersion);
        }
        client.getClientStub().setAutoVersioning(true);

        // 3. client old, server old
        client.getClientStub().setRcfRuntimeVersion(oldVersion);
        server.setRcfRuntimeVersion(oldVersion);
        BOOST_CHECK( client.echo(s0, oldVersion) == s0);
        BOOST_CHECK( client.getClientStub().getRcfRuntimeVersion() == oldVersion);
    }

    // test transparent serialization of enums
    int n = 17;
    MyEnum e = /*MyEnum::*/Two;

    RcfClient<I_EchoWithInt> clientInt(tcpEndpoint);
    RcfClient<I_EchoWithEnum> clientEnum(tcpEndpoint);

    server.bind( (I_EchoWithInt*) 0, echo);
    BOOST_CHECK( clientInt.echo(n) == n);
    BOOST_CHECK( clientEnum.echo(e) == e);

    clientEnum.getClientStub().setRcfRuntimeVersion(1);
    try {
        clientEnum.echo(e);
        BOOST_CHECK(1==0);
    }
    catch (const RCF::RemoteException &e)
    {
        BOOST_CHECK(1==1);
    }
    clientEnum.getClientStub().setRcfRuntimeVersion(currentVersion);


    server.bind( (I_EchoWithEnum*) 0, echo);
    BOOST_CHECK( clientEnum.echo(e) == e);
    BOOST_CHECK( clientInt.echo(n) == n);

    clientInt.getClientStub().setRcfRuntimeVersion(1);
    try {
        clientInt.echo(n);
        BOOST_CHECK(1==0);
    }
    catch (const RCF::RemoteException &e)
    {
        BOOST_CHECK(1==1);
    }
    clientInt.getClientStub().setRcfRuntimeVersion(currentVersion);


    return boost::exit_success;
}
