
//******************************************************************************
// RCF - Remote Call Framework
// Copyright (c) 2005 - 2007. All rights reserved.
// Consult your license for conditions of use.
// Developed by Jarl Lindrud.
// Contact: jlindrud@hotmail.com .
//******************************************************************************

#include <iostream>

#include <RCF/test/TestMinimal.hpp>

#include <RCF/Idl.hpp>
#include <RCF/RcfServer.hpp>
#include <RCF/SspiFilter.hpp>
#include <RCF/ThreadManager.hpp>
#include <RCF/Win32NamedPipeEndpoint.hpp>
#include <RCF/Win32NamedPipeServerTransport.hpp>

#include <RCF/test/ThreadGroup.hpp>
#include <RCF/test/Usernames.hpp>

class Echo
{
public:
    std::string echo(const std::string &s)
    {
        return s;
    }
    std::string echo(const std::string &s, boost::uint32_t waitMs)
    {
        Sleep(waitMs);
        return s;
    }
    std::string whatsMyUserName()
    {
        // Return clients username, by impersonating and then asking Windows
        // for the name of the currently logged on user.

        RCF::Win32NamedPipeImpersonator impersonator;
        RCF::tstring myUserName = RCF::getMyUserName();
        return std::string(myUserName.begin(), myUserName.end());
    }
};

RCF_BEGIN(I_Echo, "I_Echo")
    RCF_METHOD_R1(std::string, echo, const std::string &)
    RCF_METHOD_R2(std::string, echo, const std::string &, boost::uint32_t)
    RCF_METHOD_R0(std::string, whatsMyUserName)
RCF_END(I_Echo)

RCF::Mutex gIoMutex;

void clientTask(const RCF::tstring & pipeName, boost::uint32_t waitMs)
{
    RcfClient<I_Echo> client(( RCF::Win32NamedPipeEndpoint(pipeName)));

    boost::uint32_t t0 = RCF::getCurrentTimeMs();

    std::string s1 = "asdf";
    std::string s2 = client.echo(s1, waitMs);
    BOOST_CHECK(s1 == s2);

    boost::uint32_t t1 = RCF::getCurrentTimeMs();

    RCF::Lock lock(gIoMutex);
    std::cout << "Waited " << t1-t0 << " ms..." << std::endl;
}

void onThreadInit()
{
    RCF::Lock lock(gIoMutex);
    std::cout << "Thread initialization code here" << std::endl;    
}

void onThreadDeinit()
{
    RCF::Lock lock(gIoMutex);
    std::cout << "Thread deinitialization code here" << std::endl;
}

int test_main(int, char**)
{
    RCF::tstring pipeName = _T("TestPipe");

    Echo echo;
    RCF::RcfServer server(( RCF::Win32NamedPipeEndpoint(pipeName)));
    
    // 10 server threads
    RCF::FixedThreadPoolPtr fixedThreadPoolPtr( new RCF::FixedThreadPool(10));
    fixedThreadPoolPtr->addThreadInitFunctor( &onThreadInit);
    fixedThreadPoolPtr->addThreadDeinitFunctor( &onThreadDeinit);
    server.getServerTransportService().getTaskEntries()[0].setThreadManagerPtr( fixedThreadPoolPtr );

    server.bind<I_Echo>(echo);
    server.start();

    {
        // Make a single call, no waiting

        std::string s1 = "test";
        RcfClient<I_Echo> client(( RCF::Win32NamedPipeEndpoint(pipeName)));
        std::string s2 = client.echo(s1);
        BOOST_CHECK( s1 == s2 );
    }

    {
        // Make 5 calls, in parallel, each one waiting 2 seconds

        boost::uint32_t t0 = RCF::getCurrentTimeMs();

        boost::uint32_t waitMs = 2000;
        ThreadGroup clients;
        for (std::size_t i=0; i<5; ++i)
        {
            clients.push_back( RCF::ThreadPtr( new RCF::Thread(
                boost::bind(clientTask, pipeName, waitMs))));
        }
        joinThreadGroup(clients);

        // Total time should be little more than 2 seconds
        boost::uint32_t t1 = RCF::getCurrentTimeMs();
        std::cout << "Total time: " << t1-t0 << " ms..." << std::endl;
        BOOST_CHECK(t1-t0 < 2*waitMs);
    }

    {
        // Impersonation - using our own credentials, make a call

        RCF::tstring tUserName = RCF::getMyUserName();
        std::string userName(tUserName.begin(), tUserName.end());
        RcfClient<I_Echo> client(( RCF::Win32NamedPipeEndpoint(pipeName)));
        std::string s = client.whatsMyUserName();
        BOOST_CHECK(s == userName);
    }

    return 0;
}
